﻿/* BFx.ahk -- prep a virtual HP15C to run BRAINFUCK */
/* Goal: automatise the "how-to" recipe shown here: */
/*  https://forum.hp41.org/viewtopic.php?f=10&t=536 */
/*  As of yet the virtual HP15C is still available: */
/*  https://www.educalc.net/page/2336331/           */
/* Start: use a shortcut with target this AHK and   */
/*      in 'Start in' the directory of virtual 15C. */
/* Use: P) key p shows a warning before erasing all */
/*         programs, then loads BF interpreter.     */
/*      L) shows a file open dialog to pick a BF    */
/*         file, which is transliterated, compacted */
/*         if needed and entered if possible. Then  */
/*         all left space is converted to "tape".   */
/*     This two functions cover step 1 and 3..5 of  */
/*     the a. m. recipe -- plus, for convenience:   */
/*         Ins and F12 = same as L                  */
/*         w (west) = f-shift, e (east) = g-shift,  */
/*         Ctrl+c = copy display to clipboard       */
/*         Ctrl+v = paste value from clipboard      */
/* Quits self-acting when virual 15C is shut down.  */
/* QPL 1.0 applies, see end of file. (c) 2024 by M. */
/* "Hotel-Papa-Four-One-Charly-Xray-&-web-dot-de"   */
/* Initial version                      M. 1/16/24  */
#Requires AutoHotkey v2.0
#SingleInstance Force     ; no 2nd instance, use youngest script
SetTitleMatchMode 3       ; exact match for WinExist and alike

sid := "HP15C ahk_class #32770"  ; with variable no optimization
/* if no virtual 15C running try to start it */
if WinExist(sid)
   WinActivate
else
{
   Run("C:\Program Files\Hewlett-Packard\HP 15C Virtual Calculator\HP15C.exe", A_InitialWorkingDir)
   if NOT WinWaitActive(sid, , 2)
   {
      dc := MsgBox("Virtual 15C not up and running within 2 seconds."
             "`nAHK script ends now.", "Useless, totally", "O Iconx")
      ExitApp 9
   }
}

/* have a watchdog for HP 15C virtual Calculator */
SetTimer quitwhendue, 390

#HotIf WinActive(sid)
e::RAlt
w::
{
   if umedski()         ;  if medium skin in use
      Click 149, 433
}
p::impro()
l::
Ins::
F12::
{
   if NOT IsSet(aoo)
   {
      if MsgBox("Before loading a BRAINFUCK program`n"
       "all matrices and buffers are erased,`n"
       "also all registers will be cleared.`n`n"
       "Continue?`n"
       "Permission will be effective throughout this session."
       , "Allow or deny --", "YN Icon?") = "No"
         Return         ; quit on user request
      global aoo := 1   ; ask only once
   }
   fid := FileSelect(3,
    , "Pick a BRAINFUCK program to import"
    , "(*.bf; *.txt)")
   if fid               ; if not aborted
      loadbf(fid)
}
; ^v::pastevalue()      ;  fails due to Virtual 15C
; +v::pastevalue()      ;  simplest work around
*LCtrl::return          ;  Disable LCtrl key.
*RCtrl::return          ;  Disable RCtrl key.
#HotIf HotIfCond.CtrlV  ;  suggestion from AHK forum
*c::copydisplay()
*v::pastevalue()

#HotIf

Class HotIfCond {
   static CtrlV {
      get {
         return getKeyState("Ctrl","P")
          && WinActive(sid)
      }
   }
}
; Exit  -- end of auto-execute thread

/* * *  subroutines  * * */
quitwhendue()
{
   if not WinExist(sid)
      ExitApp 9
}

/* analyse display content and put result to clipboard */
copydisplay()
{
   if umedski()          ; only go if medium skin in use
   {
      A_Clipboard := getdisco()
      SoundBeep(371, 99)      ; indicate "done"
   }
}

/* get display content */
getdisco()
{
;   if umedski()    ; must be checked by caller
   if PixelGetColor(160, 64) = 0x000000
      disco := '-'
   else disco := ' '
   Loop 10  ; ten digits only, the sign of the mantissa is an annunciator
   {
      if PixelGetColor(A_Index * 34 + 142, 72) = 0x000000
      {
         if PixelGetColor(A_Index * 34 + 156, 72) = 0x000000
         {
            if PixelGetColor(A_Index * 34 + 157, 55) = 0x000000
            {
               if PixelGetColor(A_Index * 34 + 149, 64) = 0x000000
                  disco := disco '8'
               else disco := disco '0'
            }
            else
            {
               if PixelGetColor(A_Index * 34 + 150, 47) = 0x000000
                  disco := disco '6'
               else disco := disco 'o'
            }
         }
         else
         {
            if PixelGetColor(A_Index * 34 + 148, 79) = 0x000000
            {
               if PixelGetColor(A_Index * 34 + 157, 55) = 0x000000
                  disco := disco '2'
               else disco := disco 'E'
            }
            else
            {
               if PixelGetColor(A_Index * 34 + 157, 55) = 0x000000
                  disco := disco 'P'
               else disco := disco 'r'
            }
         }
      }
      else
      {
         if PixelGetColor(A_Index * 34 + 150, 47) = 0x000000
         {
            if PixelGetColor(A_Index * 34 + 143, 55) = 0x000000
            {
               if PixelGetColor(A_Index * 34 + 157, 55) = 0x000000
                  disco := disco '9'
               else disco := disco '5'
            }
            else
            {
               if PixelGetColor(A_Index * 34 + 148, 79) = 0x000000
                  disco := disco '3'
               else disco := disco '7'
            }
         }
         else
         {
            if PixelGetColor(A_Index * 34 + 156, 72) = 0x000000
            {
               if PixelGetColor(A_Index * 34 + 149, 64) = 0x000000
                  disco := disco '4'
               else disco := disco '1'
            }
            else
            {
               if PixelGetColor(A_Index * 34 + 149, 64) = 0x000000
               {
                  if PixelGetColor(A_Index * 34 + 157, 55) = 0x000000
                     disco := disco 'u'
                  else disco := disco '-'
               }
               else disco := disco ' '
            }
         }
      }
      if PixelGetColor(A_Index * 34 + 166, 80) = 0x000000
      {
         if PixelGetColor(A_Index * 34 + 163, 84) = 0x000000
            disco := disco ","
         else disco := disco "."
      }
   }
   return disco
}

/* paste value from clipboard */
pastevalue()
{
   if umedski()          ; only go if medium skin in use
      val2X(A_Clipboard)
}

/* push value to X-register */
val2X(valinp)
{
;   if umedski()  -- caller's task to test that
   Loop Parse valinp, "`n", "`r"
   {
      if StrLen(A_LoopField)
      {
         if RegExMatch(A_LoopField, "i)-?(\d[\., ']?)+(e-?\d{1,2})?", &r)
         {
            r := StrReplace(r[], "'")
            if InStr(r, ".", 1, , -1) < InStr(r, ",", 1, , -1)
            {
               r := StrReplace(r, ".")
               r := StrReplace(r, ",", ".")
            }
            else r := StrReplace(r, ",")
            if Ord(r) = 45
               r := SubStr(r, 2, 1) "-" SubStr(r, 3)
            r := StrReplace(r, "-", "c")
            if not (PixelGetColor(479, 96) = 0x000000)
               Send "{ESC 2}"
            Send StrReplace(r, "e", "u")
            if not (PixelGetColor(479, 96) = 0x000000)
               Send "{ESC 2}"
         }
         else SoundBeep
      }
      else SoundBeep
      break          ; use 1st record only
   }
}

/* medium skin in use? */
umedski()
{
   WinGetPos , , &W, , sid
   if (W - 806)
   {
      MsgBox "Script prepared to work with medium skin only.`nExecuting ends."
      Return 0
   }
   Return 1
}

/* load BRAINFUCK prgm */
loadbf(fid)
{
   WinActivate          ; Why?
   listing := FileRead(fid)
   if listing = ""
   {
      MsgBox "Can't read file " fid
      Return            ; quit
   }
/* process BF file line by line */
   bfc := ""
   Loop parse, listing, "`n", "`r"        ; thinly sliced
   {
      line := A_LoopField
      Switch
      {
         case InStr(line, "#", 0, 1, 1): continue
         case InStr(line, "*", 0, 1, 1): Break
         case Trim(line, " !" '"#$%&'
          "'()*/0123456789:;=?@ABCDEFGHIJKLMNOPQRSTUVWXYZ"
          "\^_` abcdefghijklmnopqrstuvwxyz{|}~") = "": continue
         Default:
         Loop parse, line, , A_Space A_Tab
         {
            Switch A_LoopField
            {
               Case ">": bfc := bfc "1"
               Case "<": bfc := bfc "2"
               Case "+": bfc := bfc "3"
               Case "-": bfc := bfc "4"
               Case ".": bfc := bfc "5"
               Case ",": bfc := bfc "6"
               Case "[": bfc := bfc "7"
               Case "]": bfc := bfc "8"
            }
         }
      }
   }
   if bfc = ""
   {
      MsgBox("No BRAINFUCK code found in file" fid)
      Return      ; quit "L" function
   }
   if not umedski()     ; which skin in use
      Return            ; not medium skin
   if NOT IsSet(bs)
      sesova()          ; set some vars
   Loop                 ; if virtual 15C is off, switch it on
   {
      if PixelSearch(&Px, &Py, 175, 46, 192, 78, 0x000000)
         Break
      if A_Index > 2
         {
            MsgBox "Unable to switch on HP 15C Virtual Calculator.`nExecuting ends."
            Return
         }
      pushkey(41)
      Sleep 80          ; wait patiently
   }
   if (PixelGetColor(479, 96) = 0x000000)
      pushkey("43 31")  ; quit PRGM mode if applicable

   rebum()              ; remove buffers and all matrices

/* get number of registers */
   pushkey("43 45")
   Sleep 35             ; wait until "running" is replaced
   mem := getdisco()
   mx := StrSplit(StrReplace(Trim(mem), '  ', " "), A_Space, , 3)
;   MsgBox("(1) " mx[1] "`n(2) " mx[2] "`nSum: " mx[1] + mx[2])
   mem := mx[1] + mx[2]
   
/* save BF code in register 3..n */
   bfc := bfc "0"
   bfcln := StrLen(bfc)
   comco := 1
   if bfcln > 17
      bfcln := Ceil(bfcln / 10)
   else comco := 0
   if bfcln > (mem - 1)
   {
      MsgBox("BF program too big, it would require`n"
       bfcln + 2 " free registers to store it.")
      Return      ; quit L process
   }
   pushkey("3 44 25")      ; register 3 is first for BF code
   Sleep 150               ; Why??
   val2X(bfcln + 2)
   pushkey("42 23 24")     ; set SIZE
   if comco
   {
      pushkey("43 4 0")    ; SF 0
      Loop bfcln
      {
         val2X(SubStr(bfc '000000000', 1, 10))
         bfc := String(SubStr(bfc, 11))
         pushkey("44 24")     ; STO (i)
         pushkey("42 6 25")   ; ISG I
      }
   }
   else
   {
      pushkey("43 5 0")    ; CF 0
      Loop parse, bfc
      {
         val2X(A_LoopField)
         pushkey("44 24")     ; STO (i)
         pushkey("42 6 25")   ; ISG I
      }
   }
   pushkey("1 42 16 1")
   val2X(mem - bfcln - 2)
   pushkey("42 23 11")
}

/* import BRAINFUCK interpreter */
impro()
{
   fid := 'BFI.15Clst'
   if not FileExist(fid)
   {
      MsgBox 'HP15C program file "' fid '" not found.'
      Return
   }
   listing := FileRead(fid)
   if listing = ""
   {
      MsgBox "Can't read file " fid
      Return            ; quit
   }

/* Virtual 15C is up -- with which skin? */
   if not umedski()
      Return
   if MsgBox("All existing programs will be discarded,`n"
    "also all matrices and buffers are erased,`n"
    "before the BRAINFUCK interpreter is loaded.`n`n"
    "Continue?", , "YN Icon?") = "No"
      Return            ; quit on user request
   WinActivate          ; else check if on or off fails
   if NOT IsSet(bs)
      sesova()          ; set some vars

/* Virtual 15C is on or off, if off switch it on */
   Loop
   {
      if PixelSearch(&Px, &Py, 175, 46, 192, 78, 0x000000)
         Break
      if A_Index > 2
         {
            MsgBox "Unable to switch on HP 15C Virtual Calculator.`nExecuting ends."
            Return
         }
      pushkey(41)
      Sleep 80          ; wait patiently
   }

/* Virtual 15C is in PRGM mode? */
   if (PixelGetColor(479, 96) = 0x000000)
      pushkey("43 31")
   rebum()              ; remove buffer and more
   pushkey("43 31")
   pushkey("42 33")     ; discard programs if any

/* in USER mode? */
   if (PixelGetColor(213, 96) = 0x000000)
      pushkey("42 45")

/* process input file line by line */
   Loop parse, listing, "`n", "`r"        ; thinly sliced
   {
      line := A_LoopField
      if InStr(line, "#!", 0, 1, 1) = 1
      {
         if InStr(line, "15C")
            Continue
         MsgBox "According to ahsh-bang not a 15C program:`n" line
         Return
      }
      if InStr(line, "*", 0, 1, 1) = 1    ; comments
         Continue
      if Trim(line) = ""                  ; .bl
         Continue
      if InStr(line, "u", 1, 4, 1) = 4    ; an extra of 15C
         togum()
      line := SubStr(line, 5, 8)
      if line = "+....1.."                ; scale
         Continue
      line := StrReplace(line, ".", "48,", 0, , 1)
      line := StrReplace(line, ",", A_Space)
      pushkey(Trim(line))                  ; missing REXX space()
      If bs
         togum()
   }
   pushkey("43 31")     ; quit program mode
}

/* remove buffers and all matrices */
rebum()
{
   pushkey("43 32")
   pushkey("43 5 8")
   pushkey("42 16 0")
}

/* Push Key or list of keys */
pushkey(lok)
{
/* no case filter required in AHK, all specials set in kmx. */
   Loop Parse, lok, A_Space
   {
      if StrLen(Trim(A_LoopField))     ; bare IF fails with string "0"
         Send kmx[Number(A_LoopField)]
   }
}

/* toggle user mode */
togum()
{
   global bs     ; read is global, write is local by default
   bs := 1 - bs  ; Bsonderer Scheiss
   pushkey("42 45")
}

/* set some vars */
sesova()
{
   global
   kmx := Map(
     11,'{F1}'     ,21,'{F7}'      ,31,'{Click 76, 353}',41,'{Esc}'
    ,12,'{F2}'     ,22,'{F8}'      ,32,'g'         ,42,'{Click 149, 433}'
    ,13,'{F3}'     ,23,'{F9}'      ,33,'{Down}'    ,43,'{RAlt}'
    ,14,'{F4}'     ,24,'{F10}'     ,34,'{Left}'    ,44,'s'
    ,15,'{F5}'     ,25,'{F11}'     ,35,'{BS}'      ,45,'r'
    ,16,'{F6}'     ,26,'u'         ,36,'{Enter}'
    ,7 ,'{Numpad7}',4 ,'{Numpad4}' ,1 ,'{Numpad1}' ,0 ,'{Numpad0}'
    ,8 ,'{Numpad8}',5 ,'{Numpad5}' ,2 ,'{Numpad2}' ,48,'{NumpadDot}'
    ,9 ,'{Numpad9}',6 ,'{Numpad6}' ,3 ,'{Numpad3}' ,49,'{+}'
    ,10,'{NumpadDiv}' ,20,'{NumpadMult}',30,'{NumpadSub}',40,'{NumpadAdd}')
   kmx.Default := 'unknown'
   bs := 0
}

/* * * * *  THE Q PUBLIC LICENSE version 1.0  * * * * *
Copyright (C) 1999 Trolltech AS, Norway.

Everyone is permitted to copy and distribute this license document. The intent of this license is to establish freedom to share and change the software regulated by this license under the open source model.

This license applies to any software containing a notice placed by the copyright holder saying that it may be distributed under the terms of the Q Public License version 1.0. Such software is herein referred to as the Software. This license covers modification and distribution of the Software, use of third-party application programs based on the Software, and development of free software which uses the Software.

Granted Rights

  1.
    You are granted the non-exclusive rights set forth in this license provided you agree to and comply with any and all conditions in this license. Whole or partial distribution of the Software, or software items that link with the Software, in any form signifies acceptance of this license.
  2.
    You may copy and distribute the Software in unmodified form provided that the entire package, including - but not restricted to - copyright, trademark notices and disclaimers, as released by the initial developer of the Software, is distributed.
  3.
    You may make modifications to the Software and distribute your modifications, in a form that is separate from the Software, such as patches. The following restrictions apply to modifications:
    a. Modifications must not alter or remove any copyright notices in the Software.
    b. When modifications to the Software are released under this license, a non-exclusive royalty-free right is granted to the initial developer of the Software to distribute your modification in future versions of the Software provided such versions remain available under these terms in addition to any other license(s) of the initial developer.
  4.
    You may distribute machine-executable forms of the Software or machine-executable forms of modified versions of the Software, provided that you meet these restrictions:
    a. You must include this license document in the distribution.
    b. You must ensure that all recipients of the machine-executable forms are also able to receive the complete machine-readable source code to the distributed Software, including all modifications, without any charge beyond the costs of data transfer, and place prominent notices in the distribution explaining this.
    c. You must ensure that all modifications included in the machine-executable forms are available under the terms of this license.
  5.
    You may use the original or modified versions of the Software to compile, link and run application programs legally developed by you or by others.
  6.
    You may develop application programs, reusable components and other software items that link with the original or modified versions of the Software. These items, when distributed, are subject to the following requirements:
    a. You must ensure that all recipients of machine-executable forms of these items are also able to receive and use the complete machine-readable source code to the items without any charge beyond the costs of data transfer.
    b. You must explicitly license all recipients of your items to use and re-distribute original and modified versions of the items in both machine-executable and source code forms. The recipients must be able to do so without any charges whatsoever, and they must be able to re-distribute to anyone they choose.
    c. If the items are not available to the general public, and the initial developer of the Software requests a copy of the items, then you must supply one.

Limitations of Liability
In no event shall the initial developers or copyright holders be liable for any damages whatsoever, including - but not restricted to - lost revenue or profits or other direct, indirect, special, incidental or consequential damages, even if they have been advised of the possibility of such damages, except to the extent invariable law, if any, provides otherwise.

No Warranty
The Software and this license document are provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

Choice of Law
This license is governed by the laws of any member state of the European Union.
* * * * *  End Of File  * * * * */